/*
   Layers.java - applet to allow display of layered data

   Classes:
      Layers, LayerCanvas, LayerInfo, LayerBox, AlphaFilter

   Edit history:
      plutchak    29-Mar-96 Cleaned up a bit
      plutchak    04-Mar-96 Fixed canvas sizing; added transparency code
      plutchak    28-Feb-96 Created (generic version)

   Author:
      Joel Plutchak, Research Programmer <plutchak@uiuc.edu>
      Department of Atmospheric Sciences
      University of Illinois at Urbana-Champaign
      105 S. Gregory Street
      Urbana IL 61801
*/
import java.applet.Applet;
import java.awt.*;
import java.awt.image.ImageObserver;
import java.awt.image.FilteredImageSource;
import java.awt.image.RGBImageFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;


import interfaces.AppletBase;
//import messages.AppEventMessage;
import messages.*;
import consts.Const;


public class Layers extends AppletBase
{
   LayerCanvas       canvas;

   BorderLayout      layout;
   GridBagLayout     gridbag;
   Label             overlayLabel, titleLabel, subtitleLabel;
   Vector            layerbox;
   CheckboxGroup     layerGroup = new CheckboxGroup();
   Button            reloadButton;
   URLButton         helpButton;
   Panel             buttonPanel, canvasPanel, titlePanel;
   Panel             reportPanel;
   Font              defaultFont, defaultBoldFont;
   Font              titleFont, subtitleFont;
   FontMetrics       defMetrics, boldMetrics;
   boolean           started = false;
   boolean 	     Master = true;

   // only the general layout should happen here-- no image loading, etc.
   public void init ( ) {
      GridBagConstraints constraints;
      Insets         in0, in1, in2, inb;
      boolean        isGroupMember, isGroupParent, isLabel, preload, addImage;
      CheckboxGroup  group;
      Vector         labels, imageInfo;
      LayerInfo      info;
      Hashtable      groupHash = new Hashtable();
      URL            url;
      String         imgSrc, label, groupName, urlString, descriptor;
      Component      component;
      Dimension      dim;
      int            index;
      char           separator;

      register(); /* collaborative registration */
      // request Master info
	byte[] tab = new byte[1];
        tab[0] = CAConsts.ISMASTER;
        CAppRequestEventMessage msg = new CAppRequestEventMessage(tab);
	send(msg);

      if (started) {   // this is for anomalous Netscape behavior
         System.out.println( "(init) already started!" );
         show();
         return;
      }

      /*
         create and place all the GUI elements:
            BorderLayout: The Applet
               West: Image Canvas Panel (canvasPanel)
               East: Image Buttons Panel (buttonPanel)
                  GridBagLayout- image load/display buttons
               North: Title Panel (titlePanel)
                  GridLayout: titleLabel and subtitleLabel
               South: Text Info (textPanel)
                  GridBagLayout:
                     reportPanel: GridLayout
      */
      canvasPanel = new Panel();
      titlePanel = new Panel();

      // instantiate some useful fonts
      defaultFont = new Font( "Helvetica",Font.PLAIN,12 );
      defaultBoldFont = new Font( "Helvetica",Font.BOLD,12 );

      /*
         read and parse applet parameters, and create GUI items
      */
      imgSrc = getParameter( "IMAGESOURCE" );

      label = getParameter("SEPARATOR");
      if ((label == null) || (label.length() == 0)) 
         separator = '|';
      else
         separator = label.charAt(0);

      if ((labels = parseParamValues( "LABELS", separator )) == null) {
         System.out.println( "Layers: A LAYERS parameter must be specified;"+
               " program aborting." );
         destroy();
         return;
      }

      // set up Panel, panel's GridBag layout, and the GridBag parameters
      buttonPanel = new Panel();
      if (defaultFont != null) buttonPanel.setFont( defaultFont );
      if (defaultBoldFont != null) buttonPanel.setFont( defaultBoldFont );

      buttonPanel.setLayout( gridbag = new GridBagLayout() );
      in0 = new Insets( 0, 4, 0, 0 );    // these offsets shouldn't be fixed
      in1 = new Insets( 0, 20, 0, 0 );
      in2 = new Insets( 0, 35, 0, 0 );
      inb = new Insets( 8, 0, 0, 0 );
      constraints = new GridBagConstraints();
      constraints.fill = GridBagConstraints.NONE; // don't resize components
      constraints.ipady = 1;                      // add a vertical buffer
      constraints.ipadx = 0;                      // no horizontal buffer
      constraints.weightx = 1.0;                  // don't clump
      constraints.gridy = GridBagConstraints.RELATIVE;    // below previous 
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      constraints.anchor = GridBagConstraints.NORTHWEST;  // upper left

      // for each label, read and parse the layer specification parameter.
      // then create the appropriate GUI component and paste it on a panel
      layerbox = new Vector();
      imageInfo = new Vector();
      for (Enumeration e = labels.elements(); e.hasMoreElements(); ) {
         label = (String) e.nextElement();      // get the label string

         if (((descriptor = getParameter( label )) == null) ||
              (descriptor == "")) {
            System.out.println( "Layers: WARNING-- label '"+label+"' has "+
                  "no matching applet parameter; label ignored." );
            continue;
         }

         // check for LABEL, GROUP, GROUP=, URL=, PRELOAD
         // String label, groupName, url;
         // boolean isGroupParent, isLabel, preload;

         // default values
         isGroupParent = false;  isGroupMember = false;
         isLabel = false;        preload = false;
         group = null;           addImage = false;

         if (getKeyValue( descriptor, "LABEL", separator ) != null)
            isLabel = true;

         if ((groupName = getKeyValue(descriptor,"GROUP",separator)) != null) {
            if (isLabel || (groupName == ""))
               isGroupParent = true;      // no URL, so it's a Group Parent
            else
               isGroupMember = true;
         }

         if ((getKeyValue( descriptor, "PRELOAD", separator )) != null)
            preload = true;

         urlString = getKeyValue( descriptor, "URL", separator );

         if (isGroupParent) {
            // create a CheckboxGroup, and a checkbox if appropriate
            group = new CheckboxGroup();
            groupHash.put( label, group );   // put in hashtable

            if (!isLabel) {           // group checkbox, not just a label
               component = new LayerBox( label, group );
               layerbox.addElement( component );
            }
            else {                    // a group parent label (not a checkbox)
               component = new Label( label );
            }

            // add component to panel in appropriate position
            constraints.insets = in1;
         }
         else if (isGroupMember) {       // group member checkbox
            // find the CheckboxGroup
            if ((group = (CheckboxGroup) groupHash.get( groupName )) == null) {
               System.out.println( "Layers: WARNING-- for descriptor '"+label+
                     "'"+ ", group '"+groupName+"' not found; label ignored.");
               continue;
            }
            component = new LayerBox( label, group, preload, preload );
            layerbox.addElement( component );
            addImage = true;

            // add component to panel in appropriate position
            constraints.insets = in2;
         }
         else if (!isLabel) {         // non-group (solo) checkbox
            component = new LayerBox( label, preload );
            layerbox.addElement( component );
            addImage = true;

            // add compnent to panel in appropriate position
            constraints.insets = in1;
         }
         else {                       // just a label
            component = new Label( label );

            // add compnent to panel in appropriate position
            constraints.insets = in0;
         }

         gridbag.setConstraints( component, constraints );
         buttonPanel.add( component );

         if (addImage) {
            AlphaFilter    filter = null;
            String         filter_s;
            Object         transp;

            if ((imgSrc != null) & (imgSrc != ""))
               urlString = imgSrc + urlString;

            // check for transparency filtering info
            if ((filter_s = getKeyValue(descriptor,"TRANSPARENCY",separator))
                  != null) {
               if ((transp = getTransparency( filter_s )) instanceof Color) {
                  filter = new AlphaFilter( (Color)transp );
               }
               else if (transp instanceof Float) {
                  int   rgb, start;

                  filter = new AlphaFilter( ((Float)transp).floatValue() );
               }
               else {   // unknown
                  System.out.println( "Layers: Warning-- can't understand "+
                        "TRANSPARENCY parameter; ignoring it" );
               }
            }

            info = new LayerInfo( label, this, urlString, filter /*,cal,
                  preload*/ );
            imageInfo.addElement( info );
         }
      }

      reloadButton = new Button( "Reload from Server" );
      constraints.insets = inb;
      constraints.weighty = 1.0;
      constraints.gridheight = 1;
      constraints.gridy = GridBagConstraints.RELATIVE;
      gridbag.setConstraints( reloadButton, constraints );
      buttonPanel.add( reloadButton );

      urlString = getParameter( "HELP" );   // see if an extra button is needed
      if (urlString != null) {
         helpButton = new URLButton(getDocumentBase(),urlString,"Help");
         constraints.gridy = GridBagConstraints.RELATIVE;
         gridbag.setConstraints( helpButton, constraints );
         buttonPanel.add( helpButton );
      }

      canvasPanel.add( canvas = new LayerCanvas( this, imageInfo ));

      // Title of applet/application
      titlePanel.setLayout( new GridLayout( 1, 2 ) );
      if ((label = getParameter( "TITLE" )) != null)
         titleLabel = new Label( label, Label.LEFT );
      else
         titleLabel = new Label( "                    ", Label.LEFT );
      titleLabel.setFont( titleFont = new Font( "TimesRoman", Font.BOLD, 24 ));
      titlePanel.add( titleLabel );
/*===
      if ((label = getParameter( "SUBTITLE" )) != null)
         subtitleLabel = new Label( label, Label.LEFT );
      else
         subtitleLabel = new Label( "", Label.RIGHT );
      subtitleLabel.setFont(
            subtitleFont = new Font( "TimesRoman", Font.BOLD, 20));
      titlePanel.add( subtitleLabel );
===*/

      // put all the components onto the window
      this.setLayout( layout = new BorderLayout( 1,1 ));
      this.add( "North", titlePanel );
      this.add( "West", canvasPanel );
      this.add( "East", buttonPanel );

      show();
      System.out.println( "(init) applet initialized" );
   }

/*
   public void stop () {
      super.stop();
   }
*/
   public void start () {
      int            n;
      String         txt;
      LayerBox       l;

      // things that should to be done after components are realized

      if (!started) {
         //boldMetrics = coordiText.getFontMetrics( defaultBoldFont );
   
         // load requested images on startup
         for (Enumeration e = layerbox.elements(); e.hasMoreElements(); ) {
   
            l = (LayerBox)e.nextElement();
            if (l.getLoadDefault()) {
               if (l.isGroupMember()) {    // also set group parent
                  LayerBox    parent;

                  for (Enumeration p = layerbox.elements();
                        p.hasMoreElements(); ) {
                     parent = (LayerBox)p.nextElement();
                     if ((parent.isGroupParent()) &&
                           (parent.group == l.group)) {
                        parent.setState( true );
                     }
                  }
               }
               canvas.resetLayer( l.getLabel(), true );
            }
         }

         started = true;
      }
      else {
         //System.out.println( "(start) already initialized" );
         show();
      }
   }

   /*
      getParameterInfo - provide short help on applet params
   */
   public String[][] getParameterInfo ( ) {
      String info[][] = {
         {"IMAGESOURCE","URL", "URL or subdirectory for image files"},
         {"LABELS","String", "Labels (short) for toggle buttons"},
         {"<label_1>","String", "Descriptors for 1st named label"},
         {"<label_2>","String", "Descriptors for 2nd named label"},
         {"<label...>","...", "..."},
         {"<label_n>","String", "Descriptors for Nth named label"},
         {"PROJECTION","String", "Map projection parameters"},
         {"NAVIGATION","String", "Class name for non-map navigation"},
         {"HELP","URL", "URL for a help document"}};

      return( info );
   }

   /*
      getAppletInfo - provide info on the applet
   */
   public String getAppletInfo () {
      return( "Layers version 0.1a, March 1, 1996\n" +
              "Author:\n" +
              "      Joel Plutchak <plutchak@uiuc.edu>\n" +
              "      Department of Atmospheric Sciences\n" +
              "      University of Illinois at Urbana-Champaign");
   }

   /*
      getTransparency - return transparency index, percentage, or RGB triplet
   */
   Object getTransparency ( String string ) {
      int      index, start;
      Object   object;

      if ((index = string.indexOf('%')) >= 0) {
         // whole-image transparency, as a percentage (100%=fully transparent)
         object = Float.valueOf( string.substring(0,index));
      }
      else if ((index = string.indexOf(',')) >= 0) { // RGB triplet
         Color    color;
         int      rgb;
         String   ss;
 
         // parse comma-separated RGB value
         rgb = 0;  
         if (string.charAt(0) == '(')    // skip parenthesis
            start = 1;
         else
            start = 0;
         for (int i = 0; i < 3; i++) {
            ss = string.substring( start, index );
            rgb += (rgb << 8) + (Integer.valueOf( ss )).intValue();
            start = index + 1;
            if ((index = string.indexOf( ',', start)) < 0) {
               index = string.length();
               if (string.charAt(index-1) == ')')
                  --index;
            }
         }
         
         object = new Color( rgb );
      }
      else {      // assume it's an index
         object = Integer.valueOf( string.substring(0,index-1));
      }
     
      return( object ); 
   }

   /*
      dispMsg - display a text message to the user

      Convenience routine; can be altered to use a TextArea or pop-up alert
   */
   public void dispMsg ( String msg ) {
      showStatus( msg );
      System.out.print( msg );
   }

   public void enableBox ( String label, boolean enable, boolean check ) {
      LayerBox   l;

      for (Enumeration e = layerbox.elements(); e.hasMoreElements(); ) {
         l = (LayerBox)e.nextElement();
         if (l.getLabel() == label) {
            if (enable) {
               l.enable();
               l.setState( check );
            }
            else {
               l.disable();
               l.setState( check );
            }
            break;
         }
      }
   }

   public void changeBox ( String label, boolean value ) {
      LayerBox   c;
      for (Enumeration e = layerbox.elements(); e.hasMoreElements(); ) {
         c = (LayerBox)e.nextElement();
	   if (label.equals(c.getLabel())) {
	    c.setState( value );
	    if (c.isGroupParent()) {
		//System.out.println ("found isGroupParent"+label+" with "+c.getLabel());
		canvas.resetLayer( ((c.getGroup()).getCurrent()).getLabel(), value);
	    }
	    else if (c.isGroupMember()) {       // group member
            	LayerBox   l;
            	boolean    disp = false;

            // some the following is a kludge due to flaw in group handling
            // (never seem to get the "deselect" events)
            	for (Enumeration e2 = layerbox.elements(); e2.hasMoreElements(); ) {
               		l = (LayerBox)e2.nextElement();
               		if (l.isGroupMember() && (l.getGroup()==c.getGroup())){
                  		if (l != c) {
                     			l.setState( !value );
                     			canvas.resetLayer( l.getLabel(), !value );
                  		}
               		}
               		else if (l.isGroupParent() && 
                     		(l.getGroup() == c.getGroup())) {
                  			disp = l.getState();
               		}
            	}
   
            // do a reset only if parent is checked
            	if (disp) {
               		canvas.resetLayer( c.getLabel(), value );
			//System.out.println ("client In disp=1\n");
		}
            }
            else {                          // "normal" checkbox
            	canvas.resetLayer( c.getLabel(), value );
		//System.out.println ("client In others\n");
            }
            break;
         }
      }
   }

   public void sendBox(String label, boolean value ) {
        String message;
	if (value) message="1"+label;
	else message="0"+label;
	byte[] t=new byte[message.length()];
	message.getBytes(0,message.length(),t,0);
         System.out.println ("Sending "+message);       
	AppEventMessage em=new AppEventMessage(Const.DATA,t);
	send(em);
   }
   // handle applet GUI events
   public boolean action ( Event ev, Object obj ) {
      int         n;
      boolean     value;
      // only master can handle events
	if (!Master) return true;

      if (ev.target instanceof LayerBox) {  // handle LayerBox
         LayerBox    c = (LayerBox)ev.target;

         value = c.getState();

         dispMsg( "" );            // clear status bar
        // System.out.println ("CheckBox label is: "+c.getLabel()+"\n"); 

         if (c.isGroupParent()) {  // group parent
	    System.out.println ("In isGroupParent()\n");
            canvas.resetLayer( ((c.getGroup()).getCurrent()).getLabel(), value);
            // sendBox(((c.getGroup()).getCurrent()).getLabel(), value);
            sendBox(c.getLabel(),value);
         }
         else if (c.isGroupMember()) {       // group member
            LayerBox   l;
            boolean    disp = false;

	    System.out.println ("In isGroupMember()\n");

            // some the following is a kludge due to flaw in group handling
            // (never seem to get the "deselect" events)
            for (Enumeration e = layerbox.elements(); e.hasMoreElements(); ) {
               l = (LayerBox)e.nextElement();
  	       //System.out.println ("check CheckBox label: "+l.getLabel()+"\n");
               if (l.isGroupMember() && (l.getGroup()==c.getGroup())){
                  if (l != c) {
		     //System.out.println ("Found l="+l.getLabel()+" c="+c.getLabel());
                     l.setState( !value );
                     canvas.resetLayer( l.getLabel(), !value );
		     sendBox( l.getLabel(), !value );
                  }
               }
               else if (l.isGroupParent() && 
                     (l.getGroup() == c.getGroup())) {
                  disp = l.getState();
               }
            }
   
            // do a reset only if parent is checked
            if (disp) {
               canvas.resetLayer( c.getLabel(), value );
	       sendBox( c.getLabel(), value );
	    }
         }
         else {                                  // "normal" checkbox
            canvas.resetLayer( c.getLabel(), value );
	    sendBox( c.getLabel(), value );
         }
      }
      else if (ev.target instanceof URLButton) {  // handle URLButtons
         URLButton   b = (URLButton)ev.target;
         URL         url;

         if ((url = b.getURL()) != null) {
            (this.getAppletContext()).showDocument( url );
         }
      }
      else if (ev.target instanceof Button) {  // handle buttons
         Button      b = (Button)ev.target;

         if (b == reloadButton) {
            // flush images so they reload from the server
            reloadLayers();
         }
      }

      return false;
   }

   /*
      reloadLayers - cause data to be fetched from the server again
   */
   void reloadLayers ( ) {
      LayerBox    l;

      canvas.flushData();        // mark all layers for reload

      // cause immediate reload of all displayed layers
      for (Enumeration e = layerbox.elements(); e.hasMoreElements(); ) {
         l = (LayerBox)e.nextElement();

         if (l.getState()) {
            if (l.isGroupParent()) {
               canvas.resetLayer(
                     ((l.getGroup()).getCurrent()).getLabel(),
                     true );
            }
            else if (!l.isGroupMember()) {
               canvas.resetLayer( l.getLabel(), true );
            }
         }
      }
   }

   /*
      getKeyValue - return value string of key/value pair

      Returns null if key not found
   */
   String getKeyValue ( String paramString, String key, char separator ) {
      int   index, next, last;

      if ((index = paramString.indexOf( key )) == -1)
         return( null );

      next = index + key.length();
      if ((next == paramString.length()) ||
            (paramString.charAt(next) == separator))
         return( "" );         // key present with no value string

      last = paramString.indexOf( separator, next+1 );  // skip '='
      if (last == -1)                    // end of string: last param in string
         return( paramString.substring( next+1 ));
      else                               // extract value up to next separator
         return( paramString.substring( next+1, last ));
   }

   /*
      parseParamValues - parse multi-valued parameter value strings
   
      Returns a Vector of strings, assuming multiple values are packed
      into the parameter string separated by a given character
   */
   Vector parseParamValues ( String keyword, char separator ) {
      Vector   v;
      String   param, value;
   
      // check for no such parameter
      if ((param = getParameter( keyword )) == null) {
         return( null );
      }
   
      // check for no parameter value (currently illegal in the HTML spec)
      if (param == "" ) {
         return( null );
      }
   
      // otherwise return vector of values assuming given separator
      v = new Vector();
      for (int i = 0; i < param.length(); ) {
         int index = param.indexOf( separator, i );
   
         if (index == -1) index = param.length();
   
         v.addElement( param.substring( i, index ));
   
         i = index + 1;
      }
   
      return( v );
   }

/** API from collaborative framework */
public void receive (AppEventMessage em) {
 
  try {
     if (em.getType() == Const.CONTROL) {
		System.out.println ("Control msg");
		CAppAnswerEventMessage msg =  (CAppAnswerEventMessage) em;
                if (msg.getAnswerType() == CAConsts.ISMASTER) {
			Master = msg.isMaster();
			if (Master) System.out.println ("I am master");
			else System.out.println ("I am slave");
		}
		return;
      }

      String message = new String(em.getData(),0,0,em.getLen());
      System.out.println ("recieving "+message);      
      boolean value = true;
      if ("0".equals(message.substring(0,1))) value = false;
      message = message.substring(1);
      System.out.println ("Label is "+message);
      changeBox(message,value);
  }
  catch (Exception e) {
    e.printStackTrace();
      }
}
}
/*
   LayerCanvas - handle image canvas painting and events
*/
class LayerCanvas extends Canvas implements ImageObserver
{
   MediaTracker      tracker;
   Vector            imgInfo;
   boolean           focus = false;
   Layers            applet;
   Dimension         maxSize;
   private Image     offScreenImage;
   private Graphics  offScreenGraphics;

   public LayerCanvas ( Layers app, Vector _imgInfo ) {
      LayerInfo   info;
      Enumeration e;
      int         n;

      applet = app;
      maxSize = minimumSize();

      tracker = new MediaTracker( this );

      imgInfo = _imgInfo;
      for (n=0, e = imgInfo.elements(); e.hasMoreElements(); ) {
         info = (LayerInfo)e.nextElement();

         if (!info.getErrorFlag()) {
            LayerInfo   ii;

            tracker.addImage( info.image, n++ );   // track image loading

            info.image.getWidth( this );    // catch height and width of images
            info.image.getHeight( this );

         }
      }
   }

   /*
      This will casue the specified image to be reloaded from the server
   */
   public void flushData ( String s ) {
      LayerInfo   img = null;

      for (Enumeration e = imgInfo.elements(); e.hasMoreElements(); ) {
         img = (LayerInfo)e.nextElement();
        
         if (img.label == s) {
            (img.image).flush();
            break;
         } 
      }
   }

   /*
      This will cause all images to be reloaded from the server
   */
   public void flushData ( ) {
      LayerInfo   img = null;

      for (Enumeration e = imgInfo.elements(); e.hasMoreElements(); ) {
         img = (LayerInfo)e.nextElement();
         
         (img.image).flush();
      }
   }

   /*
      resetLayer - causes image to be added to or removed from display

      *** Should save display state so simple additions don't do full redraw
      *** (may mess up layer ordering)
   */
   public void resetLayer ( String label, boolean state ) {
      LayerInfo   img = null;
      int         index;
      boolean     found = false;
      Enumeration e;

      for (index = 0, e = imgInfo.elements(); e.hasMoreElements(); index++) {
         img = (LayerInfo)e.nextElement();

         if (img.label == label) {
            img.setDisplayFlag( state );
            found = true;
            break;
         }
      }

      if (!found) {
         //System.out.println( "(reset) Couldn't find "+label );
         return;
      }

      // if there was a problem loading this do some cleanup
      if (tracker.isErrorID( index ) == true) {
         applet.dispMsg( "Error loading "+label+"\n" );
         System.out.println( "Error loading: "+label );
         applet.enableBox( label, false, false );
         img.setDisplayFlag( false );
         img.setErrorFlag( true );
         img.setLoadFlag( false );

         return;
      }

      // if the box was checked make sure the image gets loaded
      if (state && (tracker.statusID( index, true ) != MediaTracker.COMPLETE)) {
         applet.dispMsg( "Loading "+label+"\n" );
      }
      // else should check timeout for image reload

      repaint();
      //sendBox(label,state);
   }

   /*
      handle mouse events in the image canvas
   */

   public boolean mouseEnter ( Event e, int x, int y ) {

      focus = true;
      //requestFocus();

      return false;
   }

   public boolean mouseExit ( Event e, int x, int y ) {

      if (focus == true) {
         focus = false;
      }

      return false;
   }

   /*
      report - report mouse coordinates and satellite image value
   */
   private void report ( int x, int y ) {
      int         rgb;

   }

   /*
      mouseDown - perform action when mouse button pressed
   */
   public boolean mouseDown ( java.awt.Event event, int x, int y ) {
      int         rgb;

      if (focus == true) {
          report( x, y );         // display data to user
      }

      return false;
   }

   /*
      mouseDrag - actions to perform when mouse (button down) moved in canvas
   */
   public boolean mouseDrag ( Event event, int x, int y ) {
      int      rgb;

      if (focus == true) {
          report( x, y );         // display data to user
      }
      return false;
   }

   /*
      mouseMove - actions to perform when mouse (button up) moved in canvas
   */
   public boolean mouseMove ( Event event, int x, int y ) {
      int      rgb;

      if (focus == true) {
          report( x, y );         // display data to user
      }
      return false;
   }

   /*
      mouseUp - actions to perform when mouse button released
   */
   public boolean mouseUp ( Event event, int x, int y ) {
      int         rgb;

      if (focus == true) {
      }

      return false;
   }

   /*
      handle screen updates
   */
   public final synchronized void update ( Graphics g ) {
      Dimension   d = size();

      // make sure offscreen buffer is proper size
      if((offScreenImage == null) || (d.width != maxSize.width) ||
            (d.height != maxSize.height)) {
         offScreenImage = createImage( d.width, d.height );
         maxSize = d;
         offScreenGraphics = offScreenImage.getGraphics();
      }
      paint( offScreenGraphics );
      g.drawImage( offScreenImage, 0, 0, null );
   }


   /*
      handle image painting and reading
   */
   public void paint ( Graphics g ) {
      int         nactive;
      LayerInfo   img;

      //setBackground( Color.lightGray );
      //g.clearRect( 0, 0, maxSize.width, maxSize.height );
      g.setColor( Color.lightGray );
      g.fillRect( 0, 0, size().width, size().height );

      // make sure opaque layer gets painted first
      //for (Enumeration e = imgInfo.elements(); e.hasMoreElements(); ) {
      //   img = (LayerInfo)e.nextElement();
      //
      //   if ((img.getOpaqueFlag()) && (img.getDisplayFlag())) {
      //      g.drawImage( img.image, 0, 0, this );
      //      nactive++;
      //      break;
      //   }
      //}

      // now paint all transparent layers
      nactive = 0;
      for (Enumeration e = imgInfo.elements(); e.hasMoreElements(); ) {
         img = (LayerInfo)e.nextElement();

         if ((!img.getOpaqueFlag()) && (img.getDisplayFlag())) {
            g.drawImage( img.image, 0, 0, this );
            nactive++;
         }
      }

      // outline image display area if no layers are selected
      if (nactive == 0) {
         g.setColor( Color.black );
         g.draw3DRect( 2, 2, size().width-4, size().height-4, true );
      }
   }

   public boolean imageUpdate ( Image img, int infoflags, int x, int y,
         int width, int height ) {
      LayerInfo   infoImg;
      boolean     found;

      if ((infoflags & (HEIGHT | WIDTH)) != 0) {
         if ((infoflags & HEIGHT) != 0) {    // adjust canvas size
            if (height > maxSize.height) {
               maxSize.height = height;
               resize( maxSize );
            }
         }

         if ((infoflags & WIDTH) != 0) {    // adjust canvas size
            if (width > maxSize.width) {
               maxSize.width = width;
               resize( maxSize );
            }
         }
         //applet.layout();
      }

      if ((infoflags & (ALLBITS | ERROR)) != 0) {
         // find correct image
         for (Enumeration e = imgInfo.elements(); e.hasMoreElements(); ) {
            infoImg = (LayerInfo)e.nextElement();
   
            if (infoImg.image == img) {
               if ((infoflags & ALLBITS) != 0) {   // an image has been loaded
                  applet.enableBox( infoImg.label, true, true );
               }
               else if ((infoflags & ERROR) != 0) {
                  applet.dispMsg("Error loading data: "+infoImg.label+"\n");
                  System.out.println( "(U) Error loading: "+infoImg.label );
                  System.out.println( "    "+(infoImg.url).toExternalForm() );
                  infoImg.setErrorFlag( true );
                  infoImg.setLoadFlag( false );
                  infoImg.setDisplayFlag( false );
               }
               break;
            }
         }
         repaint();
      }

      //return (infoflags & (ALLBITS | ERROR)) == 0;
      return true;
   }

   /*
      This should return variable dimension based on image size
   */
   public Dimension minimumSize() {
      return new Dimension( 256, 256 );
   }

   /*
      This should return variable dimension based on image size
   */
   public Dimension preferredSize() {
      if (maxSize == null)
         return( minimumSize() );

      return( maxSize ) ;
   }
}

/*
   LayerInfo - encapsulates the images and associated info

   Information includes the label (used to cross-index with display buttons),
   the Image object and URL, transparency filter, calibration and navigation
   information, and state flags (whether the image is loaded, is currently
   displayed, is opaque, and/or couldn't be loaded).

*/
class LayerInfo extends Object {
   public String label;            // label for matching with checkbox
   public Image  image;            // label for matching with checkbox
   AlphaFilter   filter;           // filter for doing transparency
   URL           url;              // the URL, for reference
   Dimension     offset;           // offset from upper left of canvas (unused)
   String        calibration;      // calibration class (unused)
   String        navigation;       // navigation class (unused)
   boolean       opaqueFlag;       // a fully-opaque image (unused)
   boolean       loadFlag;         // load on startup
   boolean       dispFlag;         // currently display state
   boolean       errFlag;          // an error occurred loading the image

   // constructor
   public LayerInfo ( String l, Applet app, String urls, AlphaFilter _filter ) {
      label = l;
      calibration = null;
      navigation = null;
      filter = _filter;
      opaqueFlag = false;
      loadFlag = false;
      dispFlag = false;
      errFlag = false;

      try {
         url = new URL( app.getCodeBase(), urls );
         if (filter == null)
            image = app.getImage( url );
         else
            image = app.createImage( new FilteredImageSource(
                  (app.getImage( url )).getSource(), filter ));
      }
      catch (MalformedURLException ue) {
         System.out.println(">>> For "+urls+", bad URL: " + ue.getMessage() );
         ue.printStackTrace();
         errFlag = true;
         image = null;
      }
   }

   // data access methods
   public String getNavigation ( ) {
      return( navigation );
   }

   public String getCalibration ( ) {
      return( calibration );
   }

   public boolean getOpaqueFlag ( ) {
      return( opaqueFlag );
   }

   public boolean getLoadFlag ( ) {
      return( loadFlag );
   }

   public boolean getDisplayFlag ( ) {
      return( dispFlag );
   }

   public boolean getErrorFlag ( ) {
      return( errFlag );
   }

   public void setNavigation ( String nav ) {
      navigation = nav;
   }

   public void setCalibration ( String cal ) {
      calibration = cal;
   }

   public void setLoadFlag ( boolean b ) {
      loadFlag = b;
   }

   public void setDisplayFlag ( boolean b ) {
      dispFlag = b;
   }

   public void setErrorFlag ( boolean b ) {
      errFlag = b;
   }

   public void setOpaqueFlag ( boolean b ) {
      opaqueFlag = b;
   }

   // convenience methods
   public String toString ( ) {
      String s = "LayerInfo['"+label+"',"+opaqueFlag+","+loadFlag+","+
            dispFlag+","+errFlag+"]";

      return( s );
   }
}

/*
   LayerBox - encapsulates the image checkboxes and associated info

   Extra information is group membership, group checkbox, and whether
   to load on startup (versus strictly on demand).
*/
class LayerBox extends Checkbox {
   boolean       loadDefault;
   boolean       isGroup;
   CheckboxGroup group;

   // constructors for standard non-group checkbox
   public LayerBox ( String l ) {
      super.setLabel( l );
      isGroup = false;
      group = null;
      loadDefault = false;
   }

   public LayerBox ( String l, boolean def ) {
      super.setLabel( l );
      super.setState( def );
      isGroup = false;
      group = null;
      loadDefault = def;
   }

   // constructor for main group checkbox
   public LayerBox ( String l, CheckboxGroup g ) {
      super.setLabel( l );
      loadDefault = false;
      isGroup = true;
      group = g;
   }

   // constructors for group member checkbox
   public LayerBox ( String l, CheckboxGroup g, boolean def ) {
      super.setLabel( l );
      super.setCheckboxGroup( g );
      super.setState( def );
      loadDefault = false;
      isGroup = false;
      group = g;
   }

   public LayerBox ( String l, CheckboxGroup g, boolean def, boolean load ) {
      super.setLabel( l );
      super.setCheckboxGroup( g );
      super.setState( def );
      loadDefault = load;
      isGroup = false;
      group = g;
   }

   // methods to access variables
   public boolean isGroupMember ( ) {
      return( (group != null) && !isGroup );
   }

   public CheckboxGroup getGroup ( ) {
      return( group );
   }

   public boolean isGroupParent ( ) {
      return( isGroup );
   }

   public boolean getLoadDefault ( ) {
      return( loadDefault );
   }

   public void setLoadDefault ( boolean b ) {
      loadDefault = b;
   }

   /*
      convenience methods
   */
   public String toString ( ) {
      String s = super.toString();

      return( s );
   }
}

/*
   URLButton - adds data for a URL to go to when button is pressed
*/
class URLButton extends Button {
   URL      destination;

   public URLButton ( URL url, String string ) {
      destination = url;
      setLabel( string );
   }

   public URLButton ( URL base, String url, String string ) {
      setLabel( string );

      try {
         destination = new URL( base, url );
      }
      catch (MalformedURLException e) {
         System.out.println( "(Layers::helpButton) invalid URL: " +
               e.getMessage() );
         e.printStackTrace();
         disable();
      }
   }

   public URLButton ( String url, String string ) {
      setLabel( string );

      try {
         destination = new URL( url );
      }
      catch (MalformedURLException e) {
         System.out.println( "(Layers::helpButton) invalid URL: " +
               e.getMessage() );
         e.printStackTrace();
         disable();
      }
   }

   public URLButton ( URL url ) {
      destination = url;
      setLabel( "?" );
   }

   public URLButton ( ) {
      destination = null;
      super.setLabel( "?" );
      disable();
   }

   public URL getURL ( ) {
      return( destination );
   }

   public boolean setURL ( URL url ) {
      destination = url;
      return( true );
   }

   public boolean setURL ( String url ) {
      try {
         destination = new URL( url );
      }
      catch (MalformedURLException e) {
         System.out.println( "(Layers::URLButton) invalid URL: " +
               e.getMessage() );
         e.printStackTrace();
         disable();
         return( false );
      }
      
      return( true );
   }
}

/*
   class AlphaFilter - resets all alpha values to a specified percentage
         and/or sets a particular value (or index?) to fully transparent.
*/
class AlphaFilter extends RGBImageFilter {
   int      alphaValue;             // transparency value for whole image
   int      alphaRGB;               // color for fully-transparent pixels

   // single-color full transparency only
   public AlphaFilter ( int rgb ) {
      canFilterIndexColorModel = true;
      alphaValue = 255;
      alphaRGB = rgb;
   }

   // single-color full transparency only
   public AlphaFilter ( Color color ) {
      canFilterIndexColorModel = true;
      alphaValue = 0xFF000000;
      alphaRGB = color.getRed()<<16 + color.getGreen()<<8 + color.getBlue();
   }

   // whole-image absolute partial transp plus single-color full transp
   public AlphaFilter ( int whole, int rgb ) {
      canFilterIndexColorModel = true;
      alphaValue = whole<<24;
      alphaRGB = (rgb & 0xFFFFFF);
   }

   // whole-image partial percentage transp plus single-color full transp
   public AlphaFilter ( float whole, int rgb ) {
      canFilterIndexColorModel = true;
      alphaValue = ((int) ((100-whole)*255/100.0)) << 24;
      alphaRGB = (rgb & 0xFFFFFF);
   }

   // whole-image partial percentage transparency
   public AlphaFilter ( float whole ) {
      canFilterIndexColorModel = true;
      alphaValue = ((int) ((100-whole)*255/100.0)) << 24;

      alphaRGB = -1;
   }

   // set the alpha byte to the proper value
   public int filterRGB ( int x, int y, int rgb ) {
      if (((rgb >> 24) & 0xFF) != 0xFF) { // preserve any existing transparency
         return( rgb );
      }

      if ((alphaRGB >= 0) && ((rgb & 0xFFFFFF)==alphaRGB)) { //fully-trans pixel
         return( rgb & 0xFFFFFF );
      }

      if (alphaValue == 0xFF000000)
         return( rgb );

      return( alphaValue | (rgb & 0xFFFFFF) );  // set full-image transparency
   }
}
